import pandas as pd
from functools import reduce
static_resource_path="/ocean/projects/asc170022p/shared/Data/ukBiobank/meta_data_november_2021/"
home_directory="/ocean/projects/asc170022p/tighu/"
fs_combat_cols=['eid','53-2.0','34-0.0','31-0.0','54-2.0']

class harmonize_fs_data:

    def __init__(self):
        with open(home_directory+'/jupyter_notebooks/'+'ukb_freesurfer_folder_paths.csv') as f:
            folder_paths = f.read().splitlines()

        fs_subject_list = []
        for path in folder_paths:
            fs_subject_list.append(path.split("/")[-3])

        long_combat_df = pd.read_csv(static_resource_path + "ukb49570.csv", usecols=fs_combat_cols)
        long_combat_df = long_combat_df[long_combat_df['eid'].isin(fs_subject_list)]
        long_combat_df = long_combat_df.rename(
            columns={"eid": "Subject_ID", "31-0.0": "Sex", "34-0.0": "Year_of_Birth", "53-2.0": "Date_of_Brain_Scan",
                     "54-2.0": "Assesment_Centre"})
        long_combat_df['Assesment_Centre'] = long_combat_df['Assesment_Centre'].apply(lambda x: self.get_center(x))
        long_combat_df = long_combat_df.reset_index(drop=True)
        long_combat_df['Date_of_Brain_Scan'] = pd.to_datetime(long_combat_df['Date_of_Brain_Scan'])
        long_combat_df['Sex'] = long_combat_df['Sex'].apply(lambda x: self.ukb_subject_sex(x))
        long_combat_df['Age'] = long_combat_df['Year_of_Birth'].apply(lambda x: self.ukb_subject_age(x))
        long_combat_df['Age_at_Scan'] = long_combat_df.apply(lambda x: self.ukb_subject_age_at_scan(x), axis=1)

        vol_csv = pd.read_csv(home_directory + "ukb_aseg_stats_volume_copy.csv")
        rh_csv = pd.read_csv(home_directory + "ukb_stats_rh_thickness.csv")
        lh_csv = pd.read_csv(home_directory + "ukb_stats_lh_thickness.csv")
        df_list = [lh_csv, rh_csv, vol_csv]

        df_long_combat_merged = reduce(lambda left, right: pd.merge(left, right, on=['folder_path'],
                                                                    how='outer'), df_list)
        df_long_combat_merged['Subject_ID'] = df_long_combat_merged['folder_path'].apply(
            lambda x: self.parse_subject_id_from_path(x))
        df_long_combat_merged['Subject_ID'] = df_long_combat_merged['folder_path'].apply(
            lambda x: self.parse_subject_id_from_path(x))

        long_combat_df['Subject_ID'] = long_combat_df['Subject_ID'].astype(str)

        final_df = pd.merge(long_combat_df, df_long_combat_merged, on='Subject_ID', how='outer')
        final_df = final_df.drop('folder_path', 1)

        self.harmonize_df=final_df

    def get_input_harmonizer(self):

        return self.harmonize_df

    def get_input_harmonizer_batch(self,subject_id_list):

        long_combat_df = pd.read_csv(static_resource_path + "ukb49570.csv", usecols=fs_combat_cols)
        long_combat_df = long_combat_df[long_combat_df['eid'].isin(subject_id_list)]
        long_combat_df = long_combat_df.rename(
            columns={"eid": "Subject_ID", "31-0.0": "Sex", "34-0.0": "Year_of_Birth", "53-2.0": "Date_of_Brain_Scan",
                     "54-2.0": "Assesment_Centre"})
        long_combat_df['Assesment_Centre'] = long_combat_df['Assesment_Centre'].apply(lambda x: self.get_center(x))
        long_combat_df = long_combat_df.reset_index(drop=True)
        long_combat_df['Date_of_Brain_Scan'] = pd.to_datetime(long_combat_df['Date_of_Brain_Scan'])
        long_combat_df['Sex'] = long_combat_df['Sex'].apply(lambda x: self.ukb_subject_sex(x))
        long_combat_df['Age'] = long_combat_df['Year_of_Birth'].apply(lambda x: self.ukb_subject_age(x))
        long_combat_df['Age_at_Scan'] = long_combat_df.apply(lambda x: self.ukb_subject_age_at_scan(x), axis=1)

        fs_features_df=pd.read_csv(home_directory+"jupyter_notebooks/ukb_fs_features.csv")
        fs_features_df = fs_features_df[fs_features_df['Subject_ID'].isin(subject_id_list)]

        final_df = pd.merge(long_combat_df, fs_features_df, on='Subject_ID', how='outer')

        return final_df

    def ukb_subject_sex(self,type):
        if type == 0.0:
            return "F"

        else:
            return "M"
    def ukb_subject_age(self,year):
        return int(2022 - year)

    def ukb_subject_age_at_scan(self,row):
        return int(row['Date_of_Brain_Scan'].year - row['Year_of_Birth'])

    def parse_subject_id_from_path(self,path):
        return path.split("/")[-3]

    def get_center(self,center):
        if center == 11025.0:
            return "Cheadle"

        elif center == 11026.0:
            return "Reading"

        elif center == 11027.0:
            return "Newcastle"



from difflib import get_close_matches
static_resource_path="/ocean/projects/asc170022p/shared/Data/ukBiobank/meta_data_november_2021/"
import pandas as pd

import os
import subprocess

class bulk_data_handler:

    """
    A class to represent family of methods that can handle bulk data retrieval

    ...

    Attributes
    ----------
    columns_to_read_for_field_id : list
        list of relevant columns to read from metadata file


    Methods
    -------
    display_all_ukb_categories():
        displays all the categories

    get_field_ids_for_category(category_name):
        retrieves filed ids for a category

    get_subject_list_field_ids(category_name):
        retrieves list of subject relevant to that field id

    get_data_bulk(field_id,subject_id):
        downloads bulk data from openmind

    """


    def __init__(self):
        self.columns_to_read_for_field_id = []
        return


    def display_all_ukb_categories(self):
        """
        Utility function which lets user see all the major categories present in the UKB dataset. Reads
        categories from a static resource file inside package

        Parameters:
        Needs no parameters to operate

        Returns:
        categories list: List having all the unique categories of UKB
        :rtype: list of strings
        """

        unique_category_file_object = open(static_resource_path+"all_unique_categories.txt", "r")
        all_categories_list = unique_category_file_object.readlines()
        formatted_category_list = [a.rstrip() for a in all_categories_list]
        return formatted_category_list


    def get_field_ids_for_category(self,category_name=None):

        """
        A utility function which lets user see the associated field ids for each category, reads from a static file
        containing all the  category names and field ids

        Parameters:
        a string representing a category name

        Returns:
        categories list: pandas object having related field ids of the category
        :rtype: pandas table
        """

        temp_field_ids_df = pd.read_csv(static_resource_path+"ukb_field_ids.csv")
        return temp_field_ids_df[temp_field_ids_df['Category'] == category_name]


    def get_subject_list_field_ids(self,field_id=None):

        """
        A helper function which lets user see subjects having data related to a field_id. Reads a metadata file
        into a pandas object, only reading relevant coloumns because of size of the csv file

        Parameters:
        a int representing a field id

        Returns:
        subject list: a list of subject ids
        :rtype: list
        """

        with open(static_resource_path+'new_column_list_1.txt') as f:
            columns_list_df1 = f.readlines()

        for column_name in columns_list_df1:
            if int(column_name.split("-")[0]) == field_id:
                self.columns_to_read_for_field_id.append(column_name.strip())

        tempdf=pd.read_csv(static_resource_path+"ukb49570.csv", usecols=self.columns_to_read_for_field_id+['eid'])

        return tempdf['eid'].unique()


    def get_data_bulk(self, field_id = None, subject_id = None):
        """
        A helper function which lets user download bulk data related to a subject, invokes datalad under the hood
        download starts after user provides credentials in a new terminal shell

        Parameters:
        integers representing field id and subject id

        Returns:
        download path string: path where bulk data has been downloaded
        :rtype: str
        """

        str_subject_id=str(subject_id)
        sting_for_datalad_command=""
        for col in self.columns_to_read_for_field_id:
            modified_col=col.replace('-','_').replace('.', '_')
            temp_str=str_subject_id+"/"+str_subject_id+"_"+modified_col+".zip"+" "
            sting_for_datalad_command+=temp_str
        print(sting_for_datalad_command)


        #tempdf = pd.read_csv("/media/tighu/extended_storage/ukb49570.csv", usecols=self.columns_to_read_for_field_id+['eid'])
        #os.system("interact")
        # os.chdir('/ocean/projects/asc170022p/tighu/ukb/inputs/{}'.format(str_subject_id))
        # os.system("git config remote.origin.annex-ignore false")
        # os.chdir('/ocean/projects/asc170022p/tighu/ukb/inputs/')
        # os.system("datalad get {}".format(sting_for_datalad_command))


        return "/ocean/projects/asc170022p/tighu/ukb/inputs/"+str_subject_id


    def search_category_by_name(self , query = None):


        """
        A helper function which return closest matching category based on query given by user,
        uses built in funtion to get close match

        Parameters:
        query about the category name, probablity threshold can also be provided as an additional parameter

        Returns:
        list: a list with close matching categories name

        """

        unique_category_file_object = open("all_unique_categories.txt","r")
        all_categories_list=unique_category_file_object.readlines()
        formatted_category_list = [a.rstrip() for a in all_categories_list]
        return get_close_matches(query, formatted_category_list, 5, 0.3)

    def check_bulk_data_size(self,subject_id_list =[],field_id_list=[]):
        """
        Simple Utility function to keep tabs on the bulk data bieng fetched

        Parameters:
        subject_id_list: list with subject ids
        field_id_list: list with field ids that need to fetched for each subject

        Returns:
        bulk_files_size_estimates: Binary string of the sum of a and b
        :rtype: int
        """
        # get an estimate of bulk data files if they are needed to be fetched from openmind
        # to make sure we arent downloading excessive data from a different cluster

        return "bulk_files_size_estimates"

    def get_bulk_data_batch(self,ukb_object_list,**kwargs):
        # process the object ukb_object_list
        # make chained/parallel calls datalad calls if copy not on our cluster
        # return the a dictionary with key as subject_id+field_id and path of metadata file as value
        return "subject_dict"

    def check_location_bulk_data(self, subject_id,field_id):
        # check if we have the bulk data files in our cluster or are they only on openmind
        #set flag accordingly and return flags
        return "flag_location"

    def check_valid_subject_field_id(self,subject_id,field_id):
        #check whether the following field id exists for this subjct or not
        # if not make the user aware of the case and return flag
        return "valid_field_id_flag"



    def perform_subject_field_ids_tally(self,subject_id_list,field_id_list):
        # check to perform subject and field level tally across the two clusters and ouput how many subject and field_ids are present
        # in each

        return "subject_and_field_id_counts"
